// Copyright 2005 Mitsubishi Electric Research Laboratories All Rights Reserved.

// Permission to use, copy and modify this software and its documentation without
// fee for educational, research and non-profit purposes, is hereby granted, provided
// that the above copyright notice and the following three paragraphs appear in all copies.

// To request permission to incorporate this software into commercial products contact:
// Vice President of Marketing and Business Development;
// Mitsubishi Electric Research Laboratories (MERL), 201 Broadway, Cambridge, MA 02139 or 
// <license@merl.com>.

// IN NO EVENT SHALL MERL BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL,
// OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
// ITS DOCUMENTATION, EVEN IF MERL HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

// MERL SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED
// HEREUNDER IS ON AN "AS IS" BASIS, AND MERL HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
// UPDATES, ENHANCEMENTS OR MODIFICATIONS.

#pragma once

#include "stdlib.h"
#include "stdio.h"
#include "math.h"
#include <assert.h>
#include <iostream>


#define BRDF_SAMPLING_RES_THETA_H       90
#define BRDF_SAMPLING_RES_THETA_D       90
#define BRDF_SAMPLING_RES_PHI_D         360

// WHY is there such a scaling?
#define RED_SCALE (1.0/1500.0)
#define GREEN_SCALE (1.15/1500.0)
#define BLUE_SCALE (1.66/1500.0)
// 
// #define RED_SCALE (1.0/1500.0)
// #define GREEN_SCALE (1.0/1500.0)
// #define BLUE_SCALE (1.0/1500.0)
// #define M_PI	3.1415926535897932384626433832795

namespace MERLBRDF
{

	// dot product of two vectors
	inline double dot_product(double* v1, double* v2)
	{
		return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2];
	}

	// cross product of two vectors
	inline void cross_product (double* v1, double* v2, double* out)
	{
		out[0] = v1[1]*v2[2] - v1[2]*v2[1];
		out[1] = v1[2]*v2[0] - v1[0]*v2[2];
		out[2] = v1[0]*v2[1] - v1[1]*v2[0];
	}

	// normalize vector
	inline void normalize(double* v)
	{
		// normalize
		double len = sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
		v[0] = v[0] / len;
		v[1] = v[1] / len;
		v[2] = v[2] / len;
	}

	// rotate vector along one axis
	inline void rotate_vector(double* vector, double* axis, double angle, double* out)
	{
		double temp;
		double cross[3];
		double cos_ang = cos(angle);
		double sin_ang = sin(angle);

		out[0] = vector[0] * cos_ang;
		out[1] = vector[1] * cos_ang;
		out[2] = vector[2] * cos_ang;

		temp = axis[0]*vector[0]+axis[1]*vector[1]+axis[2]*vector[2];
		temp = temp*(1.0-cos_ang);

		out[0] += axis[0] * temp;
		out[1] += axis[1] * temp;
		out[2] += axis[2] * temp;

		cross_product (axis,vector,cross);
	
		out[0] += cross[0] * sin_ang;
		out[1] += cross[1] * sin_ang;
		out[2] += cross[2] * sin_ang;
	}


	// convert standard coordinates to half vector/difference vector coordinates
	inline void std_coords_to_half_diff_coords(double theta_in, double fi_in, double theta_out, double fi_out,
									double& theta_half,double& fi_half,double& theta_diff,double& fi_diff )
	{

		// compute in vector
		double in_vec_z = cos(theta_in);
		double proj_in_vec = sin(theta_in);
		double in_vec_x = proj_in_vec*cos(fi_in);
		double in_vec_y = proj_in_vec*sin(fi_in);
		double in[3]= {in_vec_x,in_vec_y,in_vec_z};
		normalize(in);

		// compute out vector
		double out_vec_z = cos(theta_out);
		double proj_out_vec = sin(theta_out);
		double out_vec_x = proj_out_vec*cos(fi_out);
		double out_vec_y = proj_out_vec*sin(fi_out);
		double out[3]= {out_vec_x,out_vec_y,out_vec_z};
		normalize(out);

		// compute halfway vector
		double half_x = (in_vec_x + out_vec_x)/2.0f;
		double half_y = (in_vec_y + out_vec_y)/2.0f;
		double half_z = (in_vec_z + out_vec_z)/2.0f;
		double half[3] = {half_x,half_y,half_z};
		normalize(half);

		// compute  theta_half, fi_half
		theta_half = acos(half[2]);
		fi_half = atan2(half[1], half[0]);

		double bi_normal[3] = {0.0, 1.0, 0.0};
		double normal[3] = { 0.0, 0.0, 1.0 };
		double temp[3];
		double diff[3];

		// compute diff vector
		rotate_vector(in, normal , -fi_half, temp);
		rotate_vector(temp, bi_normal, -theta_half, diff);
	
		// compute  theta_diff, fi_diff	
		if (diff[2] > 1) { diff[2] = 1; diff[1] = 0; diff[0] = 1; }
		theta_diff = acos(diff[2]);
		fi_diff = atan2(diff[1], diff[0]);

		if (isnan(theta_diff)) {
			std::cout << diff[0] << " " << diff[1] << " " << diff[2] << std::endl;
		}
		//std::cout << "In = (" << in[0] << ", " << in[1] << ", " << in[2] << ")" << std::endl;
		// std::cout << "Out = (" << out[0] << ", " << out[1] << ", " << out[2] << ")" << std::endl;
		//std::cout << "Half = (" << half[0] << ", " << half[1] << ", " << half[2] << ")" << std::endl;

	}

	inline void half_diff_coords_to_std_coords(double theta_half,double fi_half,double theta_diff,double fi_diff,
					   double& theta_in, double& fi_in, double& theta_out, double& fi_out
					  )
	{

		// compute diff vector
		double diff_vec_z = cos(theta_diff);
		double proj_diff_vec = sin(theta_diff);
		double diff_vec_x = proj_diff_vec*cos(fi_diff);
		double diff_vec_y = proj_diff_vec*sin(fi_diff);
		double diff[3] = {diff_vec_x, diff_vec_y, diff_vec_z};
		normalize(diff);

		// compute half vector
		double half_vec_z = cos(theta_half);
		double proj_half_vec = sin(theta_half);
		double half_vec_x = proj_half_vec*cos(fi_half);
		double half_vec_y = proj_half_vec*sin(fi_half);
		double half[3] = {half_vec_x, half_vec_y, half_vec_z};
		normalize(half);

		double bi_normal[3] = {0.0, 1.0, 0.0};
		double normal[3] = {0.0, 0.0, 1.0};

		double temp[3];
		double in[3];

		//compute in vector
		rotate_vector(diff, bi_normal, theta_half, temp);
		rotate_vector(temp, normal, fi_half, in);
		normalize(in);

		//compute theta_in, fi_in
		theta_in = acos(in[2]);
		fi_in = atan2(in[1], in[0]);

		//theta_out = 2.0 * theta_half - theta_in;
		
		// compute out vector
		double l = dot_product(half, in);
		double out_x = 2.0 * l * half[0] - in[0];
		double out_y = 2.0 * l * half[1] - in[1];
		double out_z = 2.0 * l * half[2] - in[2];
		double out[3] = {out_x, out_y, out_z};
		normalize(out);

		//compute theta_out, fi_out
		theta_out = acos(out[2]);
		fi_out = atan2(out[1], out[0]);

		//std::cout << "In = (" << in[0] << ", " << in[1] << ", " << in[2] << ")" << std::endl;
		//std::cout << "Half = (" << half[0] << ", " << half[1] << ", " << half[2] << ")" << std::endl;
		//std::cout << "Out = (" << out[0] << ", " << out[1] << ", " << out[2] << ")" << std::endl;
	}


	// Lookup theta_half index
	// This is a non-linear mapping!
	// In:  [0 .. pi/2]
	// Out: [0 .. 89]
	inline int theta_half_index(double theta_half)
	{
		if (theta_half <= 0.0)
			return 0;
		double theta_half_deg = ((theta_half / (M_PI/2.0))*BRDF_SAMPLING_RES_THETA_H);
		double temp = theta_half_deg*BRDF_SAMPLING_RES_THETA_H;
		temp = sqrt(temp);
		int ret_val = (int)temp;
		if (ret_val < 0) ret_val = 0;
		if (ret_val >= BRDF_SAMPLING_RES_THETA_H)
			ret_val = BRDF_SAMPLING_RES_THETA_H-1;
		return ret_val;
	}

	// Lookup theta_half from index
	// This is a non-linear mapping!
	// In:  [0 .. 89]
	// Out: [0 .. pi/2]
	inline double theta_half_from_index(int index) 
	{
	    const double temp = index;
	    const double theta_half_deg = (temp*temp) / BRDF_SAMPLING_RES_THETA_H;
		const double ret_val = theta_half_deg * (0.5*M_PI) / BRDF_SAMPLING_RES_THETA_H;
		return ret_val;
	}	

	// Lookup theta_diff index
	// In:  [0 .. pi/2]
	// Out: [0 .. 89]
	inline int theta_diff_index(double theta_diff)
	{
		int tmp = int(theta_diff / (M_PI * 0.5) * BRDF_SAMPLING_RES_THETA_D);
		if (tmp < 0)
			return 0;
		else if (tmp < BRDF_SAMPLING_RES_THETA_D - 1)
			return tmp;
		else
			return BRDF_SAMPLING_RES_THETA_D - 1;
	}

	// Lookup theta_diff from index
	// In:  [0 .. 89]
	// Out: [0 .. pi/2]
	inline double theta_diff_from_index(int index) 
	{
	    const double temp = index;
		const double theta_diff = temp * (0.5*M_PI) / BRDF_SAMPLING_RES_THETA_D;
		return theta_diff;
	}	

	// Lookup phi_diff index
	inline int phi_diff_index(double phi_diff)
	{
		// Because of reciprocity, the BRDF is unchanged under
		// phi_diff -> phi_diff + M_PI
		if (phi_diff < 0.0)
			phi_diff += M_PI;

		// In: phi_diff in [0 .. pi]
		// Out: tmp in [0 .. 179]
		int tmp = int(phi_diff / M_PI * BRDF_SAMPLING_RES_PHI_D / 2);
		if (tmp < 0)	
			return 0;
		else if (tmp < BRDF_SAMPLING_RES_PHI_D / 2 - 1)
			return tmp;
		else
			return BRDF_SAMPLING_RES_PHI_D / 2 - 1;
	}

	// Lookup phi_diff from index
	//
	inline double phi_diff_from_index(int index) 
	{
	    const double phi_diff = (index) * M_PI / (BRDF_SAMPLING_RES_PHI_D/2);
		return phi_diff;
	}

	

	// Given a pair of incoming/outgoing angles, look up the BRDF.
	inline void lookup_brdf_val(double* brdf, double theta_in, double fi_in,
				  double theta_out, double fi_out, 
				  double& red_val,double& green_val,double& blue_val)
	{
		// Convert to halfangle / difference angle coordinates
		double theta_half, fi_half, theta_diff, fi_diff;
	
		std_coords_to_half_diff_coords(theta_in, fi_in, theta_out, fi_out,
			       theta_half, fi_half, theta_diff, fi_diff);


		// Find index.
		// Note that phi_half is ignored, since isotropic BRDFs are assumed
		int ind = phi_diff_index(fi_diff) +
			  theta_diff_index(theta_diff) * BRDF_SAMPLING_RES_PHI_D / 2 +
			  theta_half_index(theta_half) * BRDF_SAMPLING_RES_PHI_D / 2 *
							 BRDF_SAMPLING_RES_THETA_D;

		red_val = brdf[ind] * RED_SCALE;
		green_val = brdf[ind + BRDF_SAMPLING_RES_THETA_H*BRDF_SAMPLING_RES_THETA_D*BRDF_SAMPLING_RES_PHI_D/2] * GREEN_SCALE;
		blue_val = brdf[ind + BRDF_SAMPLING_RES_THETA_H*BRDF_SAMPLING_RES_THETA_D*BRDF_SAMPLING_RES_PHI_D] * BLUE_SCALE;

	
		if (red_val < 0.0 || green_val < 0.0 || blue_val < 0.0)
			fprintf(stderr, "Below horizon.\n");

	}

	// Given a pair of incoming/outgoing angles, look up the BRDF.
	inline void lookup_brdf_val(double* brdf, double theta_in, double fi_in,
				  double theta_out, double fi_out, 
				  double& red_val,double& green_val,double& blue_val, 
				  double& theta_half, double& fi_half, double& theta_diff, double& fi_diff)
	{
		// Convert to halfangle / difference angle coordinates
		// double fi_diff;
	
		std_coords_to_half_diff_coords(theta_in, fi_in, theta_out, fi_out,
			       theta_half, fi_half, theta_diff, fi_diff);

#define INTERPOLATE // interpolate tri-linearly between values to get smoother function.
		// more expensive (8 access instead of 1) but smoother.
		// Especially important for grazing angles
#ifndef INTERPOLATE
		// Original code
		// Find index.
		// Note that phi_half is ignored, since isotropic BRDFs are assumed
		int ind = phi_diff_index(fi_diff) +
			  theta_diff_index(theta_diff) * BRDF_SAMPLING_RES_PHI_D / 2 +
			  theta_half_index(theta_half) * BRDF_SAMPLING_RES_PHI_D / 2 * BRDF_SAMPLING_RES_THETA_D;

		red_val = brdf[ind] * RED_SCALE;
		green_val = brdf[ind + BRDF_SAMPLING_RES_THETA_H*BRDF_SAMPLING_RES_THETA_D*BRDF_SAMPLING_RES_PHI_D/2] * GREEN_SCALE;
		blue_val = brdf[ind + BRDF_SAMPLING_RES_THETA_H*BRDF_SAMPLING_RES_THETA_D*BRDF_SAMPLING_RES_PHI_D] * BLUE_SCALE;
#else
		// Interpolate
		red_val = 0.0;
		green_val = 0.0;
		blue_val = 0.0;
		int phi_diff_i[2]; // stores phi_diff_index(fi_diff) and phi_diff_index(fi_diff) + 1
		int theta_diff_j[2]; // stores theta_diff_index and theta_diff_index + 1
		int theta_half_k[2]; // stores theta_half_index and theta_half_index + 1
		phi_diff_i[0] = phi_diff_index(fi_diff);
		phi_diff_i[1] = phi_diff_i[0] + 1;
		if (phi_diff_i[1] >= BRDF_SAMPLING_RES_PHI_D / 2) phi_diff_i[1] = phi_diff_i[0]; 
		theta_diff_j[0] = theta_diff_index(theta_diff);
		theta_diff_j[1] = theta_diff_j[0] + 1;
		if (theta_diff_j[1] >= BRDF_SAMPLING_RES_THETA_D) theta_diff_j[1] = theta_diff_j[0] ;
		theta_half_k[0] = theta_half_index(theta_half);
		theta_half_k[1] = theta_half_k[0] + 1;
		if (theta_half_k[1] >= BRDF_SAMPLING_RES_THETA_H) theta_half_k[1] = theta_half_k[0];
		if (fi_diff < 0.0) fi_diff += M_PI;
		double phi_diff_a;
		if (phi_diff_i[1] != phi_diff_i[0]) phi_diff_a = (fi_diff - phi_diff_from_index(phi_diff_i[0])) / (phi_diff_from_index(phi_diff_i[1]) - phi_diff_from_index(phi_diff_i[0]));
		else phi_diff_a = 0.5;
		if (fabs(phi_diff_a) < 1e-6) phi_diff_a = 0.0;
		assert(phi_diff_a >= 0);
		assert(phi_diff_a <= 1);
		double theta_diff_a;
		if (theta_diff_j[1] != theta_diff_j[0]) theta_diff_a = (theta_diff - theta_diff_from_index(theta_diff_j[0])) / (theta_diff_from_index(theta_diff_j[1]) - theta_diff_from_index(theta_diff_j[0]));
		else theta_diff_a = 0.5;
		if (fabs(theta_diff_a) < 1e-6) theta_diff_a = 0.0;
		assert(theta_diff_a >= 0);
		assert(theta_diff_a <= 1);
		double theta_half_a; 
		if (theta_half_k[1] != theta_half_k[0]) theta_half_a = (theta_half - theta_half_from_index(theta_half_k[0])) / (theta_half_from_index(theta_half_k[1]) - theta_half_from_index(theta_half_k[0]));
		else theta_half_a = 0.5;
		if (fabs(theta_half_a) < 1e-6) theta_half_a = 0.0;
		assert(theta_half_a >= 0);
		assert(theta_half_a <= 1);

		for (int i = 0; i < 2; i++) {
			double phi_diff_interpolant = (i == 0) ? 1. - phi_diff_a : phi_diff_a;
			for (int j = 0; j <2; j++) {
				double theta_diff_interpolant = (j == 0) ? 1. - theta_diff_a : theta_diff_a;
				for (int k = 0; k < 2; k++) {
					double theta_half_interpolant = (k == 0) ? 1. - theta_half_a : theta_half_a;
					int ind = (phi_diff_i[i]) +
						(theta_diff_j[j]) * BRDF_SAMPLING_RES_PHI_D / 2 +
						(theta_half_k[k]) * BRDF_SAMPLING_RES_PHI_D / 2 * BRDF_SAMPLING_RES_THETA_D;
					double interpolant =  phi_diff_interpolant * theta_diff_interpolant * theta_half_interpolant;

					double red_val_tmp = brdf[ind] * RED_SCALE;
					double green_val_tmp = brdf[ind + BRDF_SAMPLING_RES_THETA_H*BRDF_SAMPLING_RES_THETA_D*BRDF_SAMPLING_RES_PHI_D/2] * GREEN_SCALE;
					double blue_val_tmp = brdf[ind + BRDF_SAMPLING_RES_THETA_H*BRDF_SAMPLING_RES_THETA_D*BRDF_SAMPLING_RES_PHI_D] * BLUE_SCALE;

					if (red_val_tmp >= 0.0 && green_val_tmp >= 0.0 && blue_val_tmp >= 0.0) {
						red_val += red_val_tmp * interpolant;
						green_val += green_val_tmp * interpolant;
						blue_val += blue_val_tmp * interpolant;
					}
				}
			}
		}

#endif

	#ifdef DEBUG
		if (red_val < 0.0 || green_val < 0.0 || blue_val < 0.0)
	        fprintf(stderr, "Negative value [%f, %f, %f].\n", theta_half, theta_diff, fi_diff);
	#endif
	}


	// Read BRDF data
	inline bool read_brdf(const char *filename, double* &brdf)
	{
		FILE *f = fopen(filename, "rb");
		if (!f)
			return false;

		int dims[3];
		fread(dims, sizeof(int), 3, f);
		int n = dims[0] * dims[1] * dims[2];
		if (n != BRDF_SAMPLING_RES_THETA_H *
			 BRDF_SAMPLING_RES_THETA_D *
			 BRDF_SAMPLING_RES_PHI_D / 2) 
		{
			fprintf(stderr, "Dimensions don't match\n");
			fclose(f);
			return false;
		}

		brdf = (double*) malloc (sizeof(double)*3*n);
		fread(brdf, sizeof(double), 3*n, f);

		fclose(f);
		return true;
	}


	/*
	int main(int argc, char *argv[])
	{
		const char *filename = argv[1];
		double* brdf;

		// read brdf
		if (!read_brdf(filename, brdf)) 
		{
			fprintf(stderr, "Error reading %s\n", filename);
			exit(1);
		}

		// print out a 16x64x16x64 table table of BRDF values
		const int n = 16;
		for (int i = 0; i < n; i++) 
		{
		    double theta_in = i * 0.5 * M_PI / n;
		    for (int j = 0; j < 4*n; j++) 
			{
				double phi_in = j * 2.0 * M_PI / (4*n);
				for (int k = 0; k < n; k++) 
				{
					double theta_out = k * 0.5 * M_PI / n;
					for (int l = 0; l < 4*n; l++) 
					{
						double phi_out = l * 2.0 * M_PI / (4*n);
						double red,green,blue;
						lookup_brdf_val(brdf, theta_in, phi_in, theta_out, phi_out, red, green, blue);
						printf("%f %f %f\n", (float)red, (float)green, (float)blue);
					}
				}
		    }
		}
		return 0;
	}
	*/

} // namespace


