/**
 * Copyright (C) 2017 INRIA  http://www.inria.fr
 * Copyright (C) 2017 CNRS   http://www.cnrs.fr
 *
 * Authors:
 *  Nicolas Holzschuch:  nicolas.holzschuch@inria.fr
 *  Romain  Pacanowski:  romain.pacanowski@institutoptique.fr 
 *
 * Permission to use, copy and modify this software and its documentation without
 * fee for educational, research and non-profit purposes, is hereby granted, provided
 * that the above copyright notice appear in all copies.
 * 
 * To request permission to incorporate this software into commercial products contact
 * the authors above. 
 */

/**
 * Copyright INRIA Nicolas Holzschuch @ inria.fr  
 * Copyright CNRS  Romain Pacanowski @ institutoptique.fr 
 **/

#include <iostream>
#include <fstream>
#include <cassert>

//Version with Standard C++ 2011 Gamma Distribution
//#include <random>

#include <gsl/gsl_fit.h>
#include <gsl/gsl_statistics.h>
#include <gsl/gsl_sf.h>


// Code for computing shadowing term for the Exponential Power Distribution. 
// Two steps: first, compute integral of P22 (this code) and store into temporary file.
// Second, compute integral of the integral (see integralP2_G1.cpp), and store into header file. 
//
// Requires GNU Scientific Library (GSL)
// Compile with:  g++ -I/usr/local/include integralP2_dataGen.cpp -lm -L/usr/local/lib -lgsl
// On Ubuntu add :  -lblas
// (if GSL is installed in /usr/local, adapt to your own situation)
//
// Execute with: ./a.out > shadowingData
//
// See integralP2_G1.cpp for the next step.

double P2[100][1000]; // P2 values, indexed by kp  5./p and ks = exp(-u^(1/exponent))
double dks = 0.001;

using namespace std;

double G1_exp(double theta, double b, double p)
{
  if (theta == 0) return 1;
  // Need to compute an integral
  double t = b * tan(theta);
  double ustart = 1./t;
  double exponent = 20; 
  double ks = exp(-pow(ustart, 1./exponent));
  int kks = ceil(1000 - 1000 * ks);
  ks = 1 - kks/1000.;
  int pmin = floor(5./p);
  int pmax = ceil(5./p);
  double alpha_p = 5./p - pmin;
  

  double oldu = ustart;
  double lambda = 0;
  for (;kks < 1000; kks++) {
    double ks = 1 - kks/1000.;
    double nextks =  1 - (kks+1)/1000.;
    double u = pow(log(1/ks), exponent);
    double nextu = pow(log(1/nextks), exponent);
    if (isinf(nextu)) nextu = u;    
    double deltau = 0.5 * (nextu - oldu);
    if (isnan(deltau)) deltau = 0;
    // std::cout << u << " " << ks/dks << " " << kks << std::endl ;
    double value = alpha_p * P2[pmax-1][kks] + (1 - alpha_p) * P2[pmin-1][kks];
    lambda += deltau * (t * u - 1) * value;
    // std::cout << u << " " << ks/dks << " " << kks << " " << value << " " << deltau << " " << lambda << std::endl ;
    oldu = u;
  }
  return 1./(1. + lambda);
}


int main() {
  double dkt = 0.0001;

  for (int kp = 100; kp > 0; kp--) {
    double p = 5./kp;
    double integral2 = 0;
    double normalizationFactor = p  * gsl_sf_gammainv(1./p) / M_PI;
    
    // Version with Standard C++ 2011 Gamma Distribution
    //double normalizationFactor = p  * (1.0 / std::tgamma(1./p)) / M_PI;

    double oldu = 0;
    double exponent = 20; // 2/(p);
    for (int kks = 0; kks < 1000; kks++) {
      double ks = 1 - kks/1000.;
      double nextks =  1 - (kks+1)/1000.;
      double u = pow(log(1/ks), exponent);
      double nextu = pow(log(1/nextks), exponent);
      if (isinf(nextu)) nextu = u;
      double deltau = (nextu - oldu);
      if (isnan(deltau)) deltau = 0;
      double integral = 0; 
      double oldv = 0;
      for (double kt = 1; kt > 0; kt -= dkt) {
        double v =  pow(log(1/kt), exponent);
        double value = exp(-pow(u*u + v*v, p)); 
        double deltav = (pow(log(1/(kt-dkt)), exponent) - oldv);
        if (isnan(deltav)) deltav = 0;
        integral += deltav * value; 
        oldv = v;
      }
      integral *= normalizationFactor;
      oldu = u;
      // std::cout << "Storing " << kp << " x " << (int)(ks/dks) << " = " << integral << std::endl;
      P2[kp-1][kks] = integral;
      std::cout << integral << " ";
      // integral2 += integral * deltau;
      // std::cout << u << " " << p << " " << integral << " " << kks << " " << kp << " " << integral2 << " " << deltau << std::endl;
    }
    //std::cout << std::endl;
  //  std::cout << "# p = " << p << " integral = " << integral2 << " " << kp << std::endl;
  }

//   for (double theta = 0; theta < 0.5 * M_PI; theta += 0.01) {
     // std::cout << theta << " " << G1_exp(theta, 0.1, 0.25) << " " << G1_exp(theta, 0.01, 0.3) << std::endl;
  //}
}
